@@ -116,7 +116,7 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     self._tree_options = options
 
   def _reload_tree_parameters(self, connection, *args):
-    ""
+    "Forcibly loads tree parameters for passed in nodes from the database."
     options = self._tree_options
 
     for node in args:
@@ -172,8 +172,8 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     """
     options = self._tree_options
 
-    target, position = getattr(node, options.delayed_op_attr)
-    delattr(node, options.delayed_op_attr)
+    params, session_objs = getattr(node, options.delayed_op_attr)
+    target, position     = params
 
     self._reload_tree_parameters(connection, node, target)
 
@@ -181,7 +181,7 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
       # Easy: no target is specified, so place it as the root node of a new
       # tree. This requires just one query (to find the id of the new tree)
       # and no row updates.
-      tree_id = self._get_next_tree_id(connection)
+      tree_id = self._get_next_tree_id(connection, session_objs)
       setattr(node, options.parent_id_field.name, None)
       setattr(node, options.tree_id_field.name,   tree_id)
       setattr(node, options.left_field.name,      1)
@@ -200,7 +200,7 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
         target_tree_id = target_tree_id - 1
       else:
         node_tree_id   = target_tree_id + 1
-      self._manage_tree_gap(connection, target_tree_id, 1)
+      self._manage_tree_gap(connection, session_objs, target_tree_id, 1)
 
       setattr(node, options.parent_id_field.name, None)
       setattr(node, options.tree_id_field.name,   node_tree_id)
@@ -228,7 +228,7 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
           .where(options.pk_field == parent_id)
         ).fetchone()[0]
 
-      self._manage_position_gap(connection, tree_id, gap_target, right_shift)
+      self._manage_position_gap(connection, session_objs, tree_id, gap_target, right_shift)
 
       setattr(node, options.parent_id_field.name, parent_id)
       setattr(node, options.left_field.name,      left)
@@ -238,50 +238,37 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
 
   def after_insert(self, mapper, connection, node):
     "Just after a previously non-existent node is inserted into the tree."
-    # FIXME: this method shouldn't exist. Because sqlalchemy calls all
-    #   before_insert events before performing any updates or insertions, it's
-    #   possible for inserts to "double-up". That's what we detect and back
-    #   out of here, but really we should find a way to prevent this from ever
-    #   happening.
     options = self._tree_options
-
-    self._reload_tree_parameters(connection, node)
-
-    pk      = getattr(node, options.pk_field.name)
-    tree_id = getattr(node, options.tree_id_field.name)
-    left    = getattr(node, options.left_field.name)
-
-    count = connection.execute(sqlalchemy.select([options.pk_field])
-      .where((options.tree_id_field==tree_id) &
-             (options.left_field==left))
-      .count()
-    ).fetchone()[0]
-    if count > 1:
-      new_tree_id = self._get_next_tree_id(connection)
-      connection.execute(sqlalchemy.update(options.table)
-        .values({options.tree_id_field: new_tree_id})
-        .where(options.pk_field==pk))
-      setattr(node, options.tree_id_field.name, new_tree_id)
+    delattr(node, options.delayed_op_attr)
 
   def before_delete(self, mapper, connection, node):
     "Just prior to an existent node being deleted."
     options = self._tree_options
 
+    session_objs = getattr(node, options.delayed_op_attr)
+
     self._reload_tree_parameters(connection, node)
 
     # To prevent any cascade effects, we NULL-out any adjacency-list links to
     # the node being deleted. These links will be replaced with proper values
     # in the after_delete handler.
+    pk = getattr(node, options.pk_field.name)
     connection.execute(
       sqlalchemy.update(options.table)
         .values({options.parent_id_field: None})
-        .where(options.parent_id_field == getattr(node, options.pk_field.name))
+        .where(options.parent_id_field == pk)
     )
+    for obj in session_objs:
+      if getattr(obj,  options.parent_id_field.name) == pk:
+        setattr(obj, options.parent_id_field.name, None)
 
   def after_delete(self, mapper, connection, node):
     "Just after an existent node is updated."
     options = self._tree_options
 
+    session_objs = getattr(node, options.delayed_op_attr)
+    delattr(node, options.delayed_op_attr)
+
     parent_id = getattr(node, options.parent_id_field.name)
     tree_id   = getattr(node, options.tree_id_field.name)
     left      = getattr(node, options.left_field.name)
@@ -307,7 +294,7 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
 
       # We either close the gap formed by removing the tree (if there are no
       # children), or make room for a separate tree for each child:
-      self._manage_tree_gap(connection, tree_id, len(children)-1)
+      self._manage_tree_gap(connection, session_objs, tree_id, len(children)-1)
 
       # Now for each child we promote that child to be root node of a new
       # tree, updating all of it and its children's tree parameters:
@@ -327,8 +314,40 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
             (options.tree_id_field == tree_id) &
             (options.left_field >= child_left) &
             (options.left_field <= child_right)))
+        # Handle new nodes which are part of the subtree being promoted to
+        # root:
+        for obj in session_objs:
+          obj_tree_id = getattr(obj, options.tree_id_field.name)
+          if obj_tree_id != tree_id: continue
+          obj_left    = getattr(obj, options.left_field.name)
+          obj_right   = getattr(obj, options.right_field.name)
+          obj_depth   = getattr(obj, options.depth_field.name)
+          if obj_left >= child_left and obj_right <= child_right:
+            # Assign the new tree parameters:
+            setattr(obj, options.tree_id_field.name, next_tree_id)
+            setattr(obj, options.left_field.name,    obj_left  - shift)
+            setattr(obj, options.right_field.name,   obj_right - shift)
+            setattr(obj, options.depth_field.name,   obj_depth - 1)
         # Increment the tree_id for the next time through the loop:
         next_tree_id += 1
+      # Handle new nodes which are to be promoted to root themselves:
+      for obj in session_objs:
+        obj_tree_id = getattr(obj, options.tree_id_field.name)
+        if obj_tree_id != tree_id: continue
+        obj_left    = getattr(obj, options.left_field.name)
+        obj_right   = getattr(obj, options.right_field.name)
+        obj_depth   = getattr(obj, options.depth_field.name)
+        if obj_left > left and obj_right < right and obj_depth == depth+1:
+          # The shift is how much the positional fields (left/right) will have
+          # to be adjusted so that the new root node starts from 1.
+          shift = obj_left - 1
+          # Assign the new tree parameters:
+          setattr(obj, options.tree_id_field.name, next_tree_id)
+          setattr(obj, options.left_field.name,    obj_left  - shift)
+          setattr(obj, options.right_field.name,   obj_right - shift)
+          setattr(obj, options.depth_field.name,   obj_depth - 1)
+          # Increment the tree_id for the next time through the loop:
+          next_tree_id += 1
 
     else:
       # Child node, which is much simper than the root node case. We simply
@@ -372,6 +391,23 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
             else_ = options.depth_field)})
         # Only update the tree the original node was a part of:
         .where((options.tree_id_field == tree_id)))
+      for obj in session_objs:
+        obj_tree_id = getattr(obj, options.tree_id_field.name)
+        if obj_tree_id != tree_id: continue
+        obj_left    = getattr(obj, options.left_field.name)
+        obj_right   = getattr(obj, options.right_field.name)
+        obj_depth   = getattr(obj, options.depth_field.name)
+        if   (obj_left   > left  and
+              obj_left   < right and
+              obj_depth == depth+1):  setattr(obj, options.parent_id_field.name, parent_id)
+        if   (obj_left   > left  and
+              obj_left   < right):    setattr(obj, options.left_field.name,      obj_left  - 1)
+        elif (obj_left   > right):    setattr(obj, options.left_field.name,      obj_left  - 2)
+        if   (obj_right  > left  and
+              obj_right  < right):    setattr(obj, options.right_field.name,     obj_right - 1)
+        elif (obj_right  > right):    setattr(obj, options.right_field.name,     obj_right - 2)
+        if   (obj_left   > left  and
+              obj_left   < right):    setattr(obj, options.depth_field.name,     obj_depth - 1)
 
   def before_update(self, mapper, connection, node):
     """Called just prior to an existent node being updated.
@@ -406,7 +442,8 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     if not hasattr(node, options.delayed_op_attr):
       return
 
-    target, position = getattr(node, options.delayed_op_attr)
+    params, session_objs = getattr(node, options.delayed_op_attr)
+    target, position     = params
     delattr(node, options.delayed_op_attr)
 
     self._reload_tree_parameters(connection, node, target)
@@ -415,33 +452,41 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
 
     if target is None:
       if not node_is_root_node:
-        self._make_child_into_root_node(connection, node)
+        self._make_child_into_root_node(connection, session_objs, node)
 
     elif (getattr(target, options.left_field.name)==1 and
           position in [options.class_manager.POSITION_LEFT,
                        options.class_manager.POSITION_RIGHT]):
-      self._make_sibling_of_root_node(connection, node, target, position)
+      self._make_sibling_of_root_node(connection, session_objs, node, target, position)
 
     else:
       if node_is_root_node:
-        self._move_root_node(connection, node, target, position)
+        self._move_root_node(connection, session_objs, node, target, position)
 
       else:
-        self._move_child_node(connection, node, target, position)
+        self._move_child_node(connection, session_objs, node, target, position)
 
   def after_update(self, mapper, connection, node):
     "Just after an existent node is updated."
-    pass
+    options = self._tree_options
+    if hasattr(node, options.delayed_op_attr):
+      delattr(node, options.delayed_op_attr)
 
-  def _get_next_tree_id(self, connection):
+  def _get_next_tree_id(self, connection, session_objs):
     """Determines the next largest unused tree id."""
     options = self._tree_options
-    return connection.execute(
-      sqlalchemy.select([
-        (sqlalchemy.func.max(options.tree_id_field) + 1).label('tree_id')
-      ])).fetchone()[0] or 1
 
-  def _manage_tree_gap(self, connection, target_tree_id, size):
+    tree_ids = filter(
+      lambda tree_id:tree_id is not None, [
+        getattr(n, options.tree_id_field.name)
+        for n in session_objs])
+
+    return max(tree_ids + [connection.execute(
+      sqlalchemy.select([
+        (sqlalchemy.func.max(options.tree_id_field)).label('tree_id')
+      ])).fetchone()[0] or 0]) + 1
+
+  def _manage_tree_gap(self, connection, session_objs, target_tree_id, size):
     """Creates space for a new tree *after* the target by adding ``size`` to
     all tree id's greater than ``target_tree_id``."""
     options = self._tree_options
@@ -449,8 +494,12 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
       options.table.update()
         .values({options.tree_id_field: options.tree_id_field + size})
         .where(options.tree_id_field > target_tree_id))
+    for obj in session_objs:
+      obj_tree_id = getattr(obj, options.tree_id_field.name)
+      if obj_tree_id > target_tree_id:
+        setattr(obj, options.tree_id_field.name, obj_tree_id + size)
 
-  def _manage_position_gap(self, connection, tree_id, target, size):
+  def _manage_position_gap(self, connection, session_objs, tree_id, target, size):
     """Manages spaces in the tree identified by ``tree_id`` by changing the
     values of the left and right columns by ``size`` after the given
     ``target`` point."""
@@ -470,6 +519,15 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
           ((options.left_field  > target) |
            (options.right_field > target))
         ))
+    for obj in session_objs:
+      obj_tree_id = getattr(obj, options.tree_id_field.name)
+      if obj_tree_id != tree_id: continue
+      obj_left    = getattr(obj, options.left_field.name)
+      obj_right   = getattr(obj, options.right_field.name)
+      if obj_left  > target:
+        setattr(obj, options.left_field.name,  obj_left  + size)
+      if obj_right > target:
+        setattr(obj, options.right_field.name, obj_right + size)
 
   def _calculate_inter_tree_move_values(self, node, target, position):
     """Calculates values required when moving ``node`` relative to ``target``
@@ -513,8 +571,8 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
 
     return gap_target, depth_change, left_right_change, parent_id, right_shift
 
-  def _inter_tree_move_and_close_gap(self, connection, node, new_tree_id,
-    left_right_change, depth_change, parent_id=None):
+  def _inter_tree_move_and_close_gap(self, connection, session_objs, node,
+    new_tree_id, left_right_change, depth_change, parent_id=None):
     """Removes ``node`` from its current tree, with the given set of changes
     being applied to ``node`` and its descendants, closing the gap left by
     moving ``node`` as it does so.
@@ -525,9 +583,10 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     parent field."""
     options = self._tree_options
 
-    left  = getattr(node, options.left_field.name)
-    right = getattr(node, options.right_field.name)
-    depth = getattr(node, options.depth_field.name)
+    tree_id  = getattr(node, options.tree_id_field.name)
+    left     = getattr(node, options.left_field.name)
+    right    = getattr(node, options.right_field.name)
+    depth    = getattr(node, options.depth_field.name)
     gap_size = right - left + 1
 
     connection.execute(
@@ -551,21 +610,33 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
             [((options.left_field >= left) & (options.left_field <= right), options.depth_field + depth_change)],
             else_ = options.depth_field),
         })
-        .where(options.tree_id_field == getattr(node, options.tree_id_field.name)))
+        .where(options.tree_id_field == tree_id))
+    for obj in session_objs:
+      obj_tree_id = getattr(obj, options.tree_id_field.name)
+      if obj_tree_id != tree_id: continue
+      obj_left    = getattr(obj, options.left_field.name)
+      obj_right   = getattr(obj, options.right_field.name)
+      obj_depth   = getattr(obj, options.depth_field.name)
+      if   obj_left >= left and obj_left <= right:
+        setattr(obj, options.tree_id_field.name, new_tree_id)
+        setattr(obj, options.left_field.name,    obj_left  + left_right_change)
+        setattr(obj, options.depth_field.name,   obj_depth + depth_change)
+      elif obj_left > right:
+        setattr(obj, options.left_field.name,    obj_left  - gap_size)
+      if   obj_right >= left and obj_right <= right:
+        setattr(obj, options.right_field.name,   obj_right + left_right_change)
+      elif obj_right > right:
+        setattr(obj, options.right_field.name,   obj_right - gap_size)
+    setattr(node, options.parent_id_field.name,  parent_id)
 
-    setattr(node, options.parent_id_field.name, parent_id)
-    setattr(node, options.tree_id_field.name,   new_tree_id)
-    setattr(node, options.left_field.name,      left  + left_right_change)
-    setattr(node, options.right_field.name,     right + left_right_change)
-    setattr(node, options.depth_field.name,     depth + depth_change)
-
-  def _make_child_into_root_node(self, connection, node, new_tree_id=None):
+  def _make_child_into_root_node(self, connection, session_objs, node,
+    new_tree_id=None):
     """Removes ``node`` from its tree, making it the root node of a new tree.
     If ``new_tree_id`` is not specified a new tree id will be generated."""
     options = self._tree_options
 
     if not new_tree_id:
-      new_tree_id = self._get_next_tree_id(connection)
+      new_tree_id = self._get_next_tree_id(connection, session_objs)
 
     left  = getattr(node, options.left_field.name)
     right = getattr(node, options.right_field.name)
@@ -573,10 +644,12 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     left_right_change = 1 - left
     depth_change = -depth
 
-    self._inter_tree_move_and_close_gap(connection, node, new_tree_id,
-                                        left_right_change, depth_change)
+    self._inter_tree_move_and_close_gap(connection, session_objs, node,
+                                        new_tree_id, left_right_change,
+                                        depth_change)
 
-  def _make_sibling_of_root_node(self, connection, node, target, position):
+  def _make_sibling_of_root_node(self, connection, session_objs, node, target,
+    position):
     """Moves ``node``, making it a sibling of the given ``target`` root node
     as specified by ``position``.
 
@@ -602,15 +675,8 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
       else:
         raise ValueError(_(u"an invalid position was given: %s") % position)
 
-      self._manage_tree_gap(connection, gap_target, 1)
-
-      if tree_id > gap_target:
-        # The node's tree id has been incremented in the database -- this
-        # change must be reflected in the node object for the method call
-        # below to operate on the correct tree.
-        setattr(node, options.tree_id_field.name, tree_id + 1)
-
-      self._make_child_into_root_node(connection, node, new_tree_id)
+      self._manage_tree_gap(connection, session_objs, gap_target, 1)
+      self._make_child_into_root_node(connection, session_objs, node, new_tree_id)
 
     else:
       if position == options.class_manager.POSITION_LEFT:
@@ -649,10 +715,15 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
             (options.tree_id_field >= lower_bound) &
             (options.tree_id_field <= upper_bound)
           ))
+      for obj in session_objs:
+        obj_tree_id = getattr(obj, options.tree_id_field.name)
+        if obj_tree_id >= lower_bound and obj_tree_id <= upper_bound:
+          if obj_tree_id == tree_id:
+            setattr(obj, options.tree_id_field.name, new_tree_id)
+          else:
+            setattr(obj, options.tree_id_field.name, obj_tree_id + shift)
 
-      setattr(node, options.tree_id_field.name, new_tree_id)
-
-  def _move_root_node(self, connection, node, target, position):
+  def _move_root_node(self, connection, session_objs, node, target, position):
     """Moves root node``node`` to a different tree, inserting it relative to
     the given ``target`` node as specified by ``position``."""
     options = self._tree_options
@@ -672,7 +743,7 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
       self._calculate_inter_tree_move_values(node, target, position)
 
     # Create space for the tree which will be inserted
-    self._manage_position_gap(connection, new_tree_id, gap_target, right_shift)
+    self._manage_position_gap(connection, session_objs, new_tree_id, gap_target, right_shift)
 
     # Move the root node, making it a child node
     connection.execute(
@@ -691,24 +762,25 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
         (options.left_field    >= left)    &
         (options.left_field    <= right)
       ))
+    for obj in session_objs:
+      obj_tree_id = getattr(obj, options.tree_id_field.name)
+      if obj_tree_id != tree_id: continue
+      obj_left    = getattr(obj, options.left_field.name)
+      obj_right   = getattr(obj, options.right_field.name)
+      obj_depth   = getattr(obj, options.depth_field.name)
+      if obj_left >= left and obj_left <= right:
+        setattr(obj, options.tree_id_field.name, new_tree_id)
+        setattr(obj, options.left_field.name,    obj_left  + left_right_change)
+        setattr(obj, options.right_field.name,   obj_right + left_right_change)
+        setattr(obj, options.depth_field.name,   obj_depth + depth_change)
+    # Update the former root node to be consistent with the updated
+    # tree in the database:
+    setattr(node, options.parent_id_field.name, parent_id)
 
     # Remove the gap previously occupied by the tree
-    self._manage_tree_gap(connection, tree_id - 1, -1)
+    self._manage_tree_gap(connection, session_objs, tree_id - 1, -1)
 
-    # Closing the tree id gap might have affected the new tree id. We detect
-    # this so we don't end up with an inconsistent structure later:
-    if new_tree_id >= tree_id:
-      new_tree_id -= 1
-
-    # Update the former root node to be consistent with the updated
-    # tree in the database.
-    setattr(node, options.parent_id_field.name, parent_id)
-    setattr(node, options.tree_id_field.name,   new_tree_id)
-    setattr(node, options.left_field.name,      left  + left_right_change)
-    setattr(node, options.right_field.name,     right + left_right_change)
-    setattr(node, options.depth_field.name,     depth + depth_change)
-
-  def _move_child_node(self, connection, node, target, position):
+  def _move_child_node(self, connection, session_objs, node, target, position):
     """Calls the appropriate method to move child node ``node`` relative to
     the given ``target`` node as specified by ``position``."""
     options = self._tree_options
@@ -717,11 +789,11 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     new_tree_id = getattr(target, options.tree_id_field.name)
 
     if tree_id == new_tree_id:
-      self._move_child_within_tree(connection, node, target, position)
+      self._move_child_within_tree(connection, session_objs, node, target, position)
     else:
-      self._move_child_to_new_tree(connection, node, target, position)
+      self._move_child_to_new_tree(connection, session_objs, node, target, position)
 
-  def _move_child_to_new_tree(self, connection, node, target, position):
+  def _move_child_to_new_tree(self, connection, session_objs, node, target, position):
     """Moves child node ``node`` to a different tree, inserting it relative to
     the given ``target`` node in the new tree as specified by ``position``."""
     options = self._tree_options
@@ -736,13 +808,13 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
       self._calculate_inter_tree_move_values(node, target, position)
 
     # Make space for the subtree which will be moved
-    self._manage_position_gap(connection, new_tree_id, gap_target, width)
+    self._manage_position_gap(connection, session_objs, new_tree_id, gap_target, width)
 
     # Move the subtree
-    self._inter_tree_move_and_close_gap(connection, node, new_tree_id,
+    self._inter_tree_move_and_close_gap(connection, session_objs, node, new_tree_id,
       left_right_change, depth_change, parent_id)
 
-  def _move_child_within_tree(self, connection, node, target, position):
+  def _move_child_within_tree(self, connection, session_objs, node, target, position):
     """Moves child node ``node`` within its current tree relative to the given
     ``target`` node as specified by ``position``."""
     options = self._tree_options
@@ -817,9 +889,9 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
     connection.execute(
       options.table.update()
       .values({
-        options.depth_field:     sqlalchemy.case(
-          [((options.left_field >= left) & (options.left_field <= right), options.depth_field + depth_change)],
-          else_ = options.depth_field),
+        options.parent_id_field: sqlalchemy.case(
+          [(options.pk_field == getattr(node, options.pk_field.name), parent_id)],
+          else_ = options.parent_id_field),
         options.left_field:      sqlalchemy.case(
           [((options.left_field >= left)          & (options.left_field <= right),          options.left_field + left_right_change),
            ((options.left_field >= left_boundary) & (options.left_field <= right_boundary), options.left_field + gap_size)],
@@ -828,17 +900,28 @@ class TreeMapperExtension(sqlalchemy.orm.interfaces.MapperExtension):
           [((options.right_field >= left)          & (options.right_field <= right),          options.right_field + left_right_change),
            ((options.right_field >= left_boundary) & (options.right_field <= right_boundary), options.right_field + gap_size)],
           else_ = options.right_field),
-        options.parent_id_field: sqlalchemy.case(
-          [(options.pk_field == getattr(node, options.pk_field.name), parent_id)],
-          else_ = options.parent_id_field),
+        options.depth_field:     sqlalchemy.case(
+          [((options.left_field >= left) & (options.left_field <= right), options.depth_field + depth_change)],
+          else_ = options.depth_field),
       })
       .where(options.tree_id_field == tree_id))
-
+    for obj in session_objs:
+      obj_tree_id = getattr(obj, options.tree_id_field.name)
+      if obj_tree_id != tree_id: continue
+      obj_left    = getattr(obj, options.left_field.name)
+      obj_right   = getattr(obj, options.right_field.name)
+      obj_depth   = getattr(obj, options.depth_field.name)
+      if   obj_left  >= left          and obj_left  <= right:
+        setattr(obj, options.left_field.name,  obj_left  + left_right_change)
+        setattr(obj, options.depth_field.name, obj_depth + depth_change)
+      elif obj_left  >= left_boundary and obj_left  <= right_boundary:
+        setattr(obj, options.left_field.name,  obj_left  + gap_size)
+      if   obj_right >= left          and obj_right <= right:
+        setattr(obj, options.right_field.name, obj_right + left_right_change)
+      elif obj_right >= left_boundary and obj_right <= right_boundary:
+        setattr(obj, options.right_field.name, obj_right + gap_size)
     # Update the node object to be consistent database.
     setattr(node, options.parent_id_field.name, parent_id)
-    setattr(node, options.left_field.name,      new_left)
-    setattr(node, options.right_field.name,     new_right)
-    setattr(node, options.depth_field.name,     depth + depth_change)
 
 # ===----------------------------------------------------------------------===
 
@@ -862,12 +945,17 @@ class TreeSessionExtension(sqlalchemy.orm.interfaces.SessionExtension):
     "Just prior to a flush event, while we still have time to modify the flush plan."
     options = self._tree_options
 
+    session_objs = filter(
+      lambda n:isinstance(n, options.node_class),
+      session.new.union(session.identity_map.values()))
+
     for node in session.new.union(session.dirty):
       if not isinstance(node, self._node_class):
         continue
 
       if hasattr(node, options.delayed_op_attr):
-        pass
+        setattr(node, options.delayed_op_attr,
+          (getattr(node, options.delayed_op_attr), session_objs))
 
       elif (node in session.new or
             sqlalchemy.orm.attributes.get_history(
@@ -881,7 +969,10 @@ class TreeSessionExtension(sqlalchemy.orm.interfaces.SessionExtension):
           position = options.class_manager.POSITION_LAST_CHILD
           target   = getattr(node, options.parent_field_name)
 
-        setattr(node, options.delayed_op_attr, (target, position))
+        setattr(node, options.delayed_op_attr, ((target, position), session_objs))
+
+    for node in filter(lambda n:isinstance(n, options.node_class), session.deleted):
+      setattr(node, options.delayed_op_attr, session_objs)
 
 # ===----------------------------------------------------------------------===
 # End of File
